; bm_ser - serial input module for BLINKENmini
; version 1.0 date 2006-01-18
; Copyright (C) 2005-2006 Stefan Schuermans <1stein@schuermans.info>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

; clock frequency: 12 MHz (external crystal)

; PA0..1: unused, reserved for crystal
; PA2: reset
; PB0..7: data bus: output to 74HC4094s, input from SNESs
; PD0: serial input (RX)
; PD1: serial output (TX)
; PD2..4: unused
; PD5: strobe: latch 74HC4094s, load SNESs
; PD6: clock: 74HC4094s, SNESs

; serial protocol: 115200, 8N1
;   input:
;     message-format: MCUF 18x8-1/256
;     message: <header> <data>
;     header: <magic> <height> <width> <channels> <maxval>
;     magic: 0x23 0x54 0x26 0x66
;     height: 0x00 0x08
;     width: 0x00 0x12
;     channels: 0x00 0x01
;     maxval: 0x00 0xFF
;     data: <line0> ... <line7>
;     line: <pixel0> ... <pixel17>
;     pixel: 0x00..0xFF
;   output:
;     message: (<snes0>|(0x80 0x00)) <snes1> <snes2> <snes3>
;     snes (bits): 0 0 <R> <L> <X> <A> <right> <left>   0 0 <down> <up> <start> <select> <Y> <B>



.INCLUDE        "tn2313def.inc"



; IO pins
;  - data port is input by default, it is only switched to output for a short time when neccessary
.equ    PORT_CLK        =       PORTD           ; clock: PD6
.equ    BIT_CLK         =       6
.equ    PORT_STR        =       PORTD           ; strobe: PD5
.equ    BIT_STR         =       5
.equ    PORT_DATA       =       PORTB           ; data bus: PB
.equ    PIN_DATA        =       PINB
.equ    DDR_DATA        =       DDRB



; register X is reserved for usage in interrupt !!!



; general purpose registers for interrupt
.def    I_TMP           =       r16
.def    I_CNT           =       r17
.def    I_REG0          =       r0
.def    I_REG1          =       r1
.def    I_REG2          =       r2

; status register backup in interrupt
.def    I_SREG          =       r3



; general purpose registers
.def    TMP             =       r18
.def    CNT             =       r19
.def    REG0            =       r4
.def    REG1            =       r5
.def    REG2            =       r6



; current greyscale step
;  - counts 49 steps (0..48) to output 8 quadratic greyscale values
;  - for i = 0..6:
;     - in step i^2: output black/white picture, turn on pixel with value > i
;  - in step 42: read SNES keys (1st half)
;  - in step 46: read SNES keys (2nd half)
.def    STEP            =       r20

; addresses of output buffers to use (see also OUT_BUF_*)
;  - flag to indicate assignment of output buffers
;    - 0         output buffer 1 for output, output buffer 2 for receiving
;    - 1 (!= 0): output buffer 2 for output, output buffer 1 for receiving
.def    OUT_PTR_FLAG    =       r21
;  - pointer to end of output buffer (used for output to LEDs)
.def    OUT_PTR_OUT_E_L =       r7
.def    OUT_PTR_OUT_E_H =       r8
;  - pointer to receive buffer (used for receiving new values)
.def    OUT_PTR_RECV_L  =       r9
.def    OUT_PTR_RECV_H  =       r10

; flag to indicate that new SNES key data to be sent to PC is available
;  - bit SNES_F_DATA
;    - 0: no new SNES key data available,
;         interrupt may fetch new data and set this flag
;    - 1: new SNES key data is available
;         main loop may sent data and reset flag
;  - bit SNES_F_2ND
;    - set by interrupt to indicate that 2nd half of work must be done
.def    SNES_FLAG       =       r22
.equ    SNES_F_DATA     =       0
.equ    SNES_F_2ND      =       1

; receive step
;  - counts bytes received so far
.def    RECV_STEP       =       r23

; receive coordinates
;  - x coordinate of next pixel times 3 (3 * x_coord)
.def    RECV_X_3        =       r11
;  - y coordinate of next pixel as bitmask (1 << y_coord)
.def    RECV_Y_MASK     =       r12

; transmit step
;  - counts bytes to send
.def    TRANS_STEP      =       r24



.DSEG
.ORG    0x060

; output buffers with values to output (see also OUT_PTR_*)
;  - every pixel is stored as 3 bits distributed over 3 bytes
;  - the bit b ([0..2]) of the value of the pixel at x,y ([0..17],[0..7])
;    is stored in bit y of byte x*3+b
OUT_BUF_1:      .BYTE   18*3
OUT_BUF_2:      .BYTE   18*3

; key states read from SNES pads
;  - key k ([0..11]) of pad p ([0..3]) is stored in bit p+(k%2)*4 of byte k/2
SNES_BUF:       .BYTE   6



.CSEG
.ORG    0x000
        rjmp    ENTRY                           ; RESET
        reti                                    ; INT0
        reti                                    ; INT1
        reti                                    ; TIMER1_CAPT
        reti                                    ; TIMER1_COMPA
        reti                                    ; TIMER1_OVF
        reti                                    ; TIMER0_OVF
        reti                                    ; USART0_RX
        reti                                    ; USART0_UDRE
        reti                                    ; USART0_TX
        reti                                    ; ANALOG_COMP
        reti                                    ; PC_INT
        reti                                    ; TIMER1_COMPB
        rjmp    INT_TIMER                       ; TIMER0_COMPA
        reti                                    ; TIMER0_COMPB
        reti                                    ; USI_START
        reti                                    ; USI_OVERFLOW
        reti                                    ; EE_READY
        reti                                    ; WDT



; code entry point
ENTRY:
; initialize stack pointer
        ldi     TMP,RAMEND
        out     SPL,TMP
; set clock prescaler to 1:1
        ldi     TMP,1<<CLKPCE
        out     CLKPR,TMP
        ldi     TMP,0
        out     CLKPR,TMP
; set ports A and D to output (except PD0/RXD) and port B to input
        ldi     TMP,0x00
        out     PORTA,TMP
        out     PORTB,TMP
        out     PORTD,TMP
        ldi     TMP,0xFF
        out     DDRA,TMP
        ldi     TMP,0x00
        out     DDRB,TMP
        ldi     TMP,0xFE
        out     DDRD,TMP
; disable anlaog comparator
        ldi     TMP,0x80
        out     ACSR,TMP
; enable usart
        ldi     TMP,1<<U2X                      ; double speed
        out     UCSRA,TMP
        ldi     TMP,1<<RXEN|1<<TXEN             ; receiver and transmitter enabled
        out     UCSRB,TMP
        ldi     TMP,1<<UCSZ1|1<<UCSZ0           ; asynchronous, 8N1: 8 data bits, no parity, 1 stop bit
        out     UCSRC,TMP
        ldi     TMP,0                           ; speed = Fosc / (8 * (UBRR + 1))
        out     UBRRH,TMP                       ;       = 12MHz / (8 * (12 + 1))
        ldi     TMP,12                          ;       = 115384bps (115200bps, 0.16% deviation)
        out     UBRRL,TMP
; set up timer 0 (interrupt timer)
        ldi     TMP,1<<WGM01                    ; clear timer on compare match
        out     TCCR0A,TMP
        ldi     TMP,1<<CS01|1<<CS00             ; use system clock with prescaler 1:64 (5.333us)
        out     TCCR0B,TMP
        ldi     TMP,36                          ; cycle time 36 timer steps (192us)
        out     OCR0A,TMP
        ldi     TMP,1<<OCIE0A                   ; timer 0 compare match interrupt enabled
        out     TIMSK,TMP
; set up timer 1 (timeout timer)
        ldi     TMP,0                           ; clear timer on compare match
        out     TCCR1A,TMP
        ldi     TMP,1<<CS12|1<<CS10|1<<WGM12    ; use system clock with prescaler 1:1024 (85.333us)
        out     TCCR1B,TMP
        ldi     TMP,0                           ; no output compare
        out     TCCR1C,TMP
        ldi     TMP,228                         ; timer steps 58594 (cycle time 5s)
        out     OCR1AH,TMP
        ldi     TMP,226
        out     OCR1AL,TMP
; enable watchdog (64ms)
        wdr
        ldi     TMP,1<<WDCE|1<<WDE
        out     WDTCR,TMP
        ldi     TMP,1<<WDE|1<<WDP1
        out     WDTCR,TMP
        wdr
; jump to main program
        rjmp    MAIN



; output a black/white picture
;  - turn on all pixel with value >= 1
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_1:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ldi     I_TMP,0xFF                      ; data port to output
        out     DDR_DATA,I_TMP
        ldi     I_CNT,18                        ; 18 columns
OUT_BW_1_LOOP:
        ld      I_REG2,-X                       ; get bits of column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 001: D2 | D1 | D0
        or      I_REG2,I_REG0
        out     PORT_DATA,I_REG2                ; output data
        nop
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next column
        brne    OUT_BW_1_LOOP
        ldi     I_TMP,0x00                      ; data port to input
        out     DDR_DATA,I_TMP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 2
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_2:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ldi     I_TMP,0xFF                      ; data port to output
        out     DDR_DATA,I_TMP
        ldi     I_CNT,18                        ; 18 columns
OUT_BW_2_LOOP:
        ld      I_REG2,-X                       ; get bits of column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 010: D2 | D1
        out     PORT_DATA,I_REG2                ; output data
        nop
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next column
        brne    OUT_BW_2_LOOP
        ldi     I_TMP,0x00                      ; data port to input
        out     DDR_DATA,I_TMP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 3
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_3:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ldi     I_TMP,0xFF                      ; data port to output
        out     DDR_DATA,I_TMP
        ldi     I_CNT,18                        ; 18 columns
OUT_BW_3_LOOP:
        ld      I_REG2,-X                       ; get bits of column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG1,I_REG0                   ; >= 011: D2 | (D1 & D0)
        or      I_REG2,I_REG1
        out     PORT_DATA,I_REG2                ; output data
        nop
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next column
        brne    OUT_BW_3_LOOP
        ldi     I_TMP,0x00                      ; data port to input
        out     DDR_DATA,I_TMP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 4
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_4:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ldi     I_TMP,0xFF                      ; data port to output
        out     DDR_DATA,I_TMP
        ldi     I_CNT,18                        ; 18 columns
OUT_BW_4_LOOP:
        ld      I_REG2,-X                       ; get bits of column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
                                                ; >= 100: D2
        out     PORT_DATA,I_REG2                ; output data
        nop
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next column
        brne    OUT_BW_4_LOOP
        ldi     I_TMP,0x00                      ; data port to input
        out     DDR_DATA,I_TMP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 5
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_5:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ldi     I_TMP,0xFF                      ; data port to output
        out     DDR_DATA,I_TMP
        ldi     I_CNT,18                        ; 18 columns
OUT_BW_5_LOOP:
        ld      I_REG2,-X                       ; get bits of column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG1,I_REG0                   ; >= 101: D2 & (D1 | D0)
        and     I_REG2,I_REG1
        out     PORT_DATA,I_REG2                ; output data
        nop
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next column
        brne    OUT_BW_5_LOOP
        ldi     I_TMP,0x00                      ; data port to input
        out     DDR_DATA,I_TMP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 6
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_6:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ldi     I_TMP,0xFF                      ; data port to output
        out     DDR_DATA,I_TMP
        ldi     I_CNT,18                        ; 18 columns
OUT_BW_6_LOOP:
        ld      I_REG2,-X                       ; get bits of column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 110: D2 & D1
        out     PORT_DATA,I_REG2                ; output data
        nop
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next column
        brne    OUT_BW_6_LOOP
        ldi     I_TMP,0x00                      ; data port to input
        out     DDR_DATA,I_TMP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 7
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_7:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ldi     I_TMP,0xFF                      ; data port to output
        out     DDR_DATA,I_TMP
        ldi     I_CNT,18                        ; 18 columns
OUT_BW_7_LOOP:
        ld      I_REG2,-X                       ; get bits of column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 111: D2 & D1 & D0
        and     I_REG2,I_REG0
        out     PORT_DATA,I_REG2                ; output data
        nop
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next column
        brne    OUT_BW_7_LOOP
        ldi     I_TMP,0x00                      ; data port to input
        out     DDR_DATA,I_TMP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret



; output a clock impulse and delay to give SNES pads some time
;  - used: I_TMP
;  - used stack depth (without call to this function): 0
SNES_CLK_DELAY:
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse
        nop
        cbi     PORT_CLK,BIT_CLK
        ldi     I_TMP,8
SNES_CLK_DELAY_LOOP:
        dec     I_TMP
        brne    SNES_CLK_DELAY_LOOP
        ret



; read SNES keys (1st half)
;  - uses: X, SNES_FLAG, I_CNT, I_TMP, I_REG0
;  - used stack depth (without call to this function): 2
IN_SNES_1:
        sbrc    SNES_FLAG,SNES_F_DATA           ; fetch new data only if SNES flag is cleared
        rjmp    IN_SNES_1_END
        ldi     XH,high(SNES_BUF)               ; begin at begin of SNES key states
        ldi     XL,low(SNES_BUF)
                                                ; strobe impulse was already output during last output of b/w pic 7
        ldi     I_CNT,3                         ; 3 times 2 keys
IN_SNES_1_LOOP:
        in      I_TMP,PIN_DATA                  ; read SNES key state of 1st key into low part of I_REG0
        andi    I_TMP,0x0F
        mov     I_REG0,I_TMP
        rcall   SNES_CLK_DELAY
        in      I_TMP,PIN_DATA                  ; read SNES key state of 2nd key into high part of I_REG0
        andi    I_TMP,0x0F
        swap    I_TMP
        or      I_REG0,I_TMP
        st      X+,I_REG0                       ; store SNES key data of both keys
        rcall   SNES_CLK_DELAY
        dec     I_CNT                           ; next column
        brne    IN_SNES_1_LOOP
        sbr     SNES_FLAG,1<<SNES_F_2ND         ; schedule 2nd part of work
IN_SNES_1_END:
        ret



; read SNES keys (2nd half)
;  - uses: X, SNES_FLAG, I_CNT, I_TMP, I_REG0
;  - used stack depth (without call to this function): 2
IN_SNES_2:
        sbrs    SNES_FLAG,SNES_F_2ND            ; do 2nd part of work only if requested (by 1st half)
        rjmp    IN_SNES_2_END
        ldi     XH,high(SNES_BUF+3)             ; begin at middle of SNES key states
        ldi     XL,low(SNES_BUF+3)
        ldi     I_CNT,3                         ; 3 times 2 keys
IN_SNES_2_LOOP:
        in      I_TMP,PIN_DATA                  ; read SNES key state of 1st key into low part of I_REG0
        andi    I_TMP,0x0F
        mov     I_REG0,I_TMP
        rcall   SNES_CLK_DELAY
        in      I_TMP,PIN_DATA                  ; read SNES key state of 2nd key into high part of I_REG0
        andi    I_TMP,0x0F
        swap    I_TMP
        or      I_REG0,I_TMP
        st      X+,I_REG0                       ; store SNES key data of both keys
        rcall   SNES_CLK_DELAY
        dec     I_CNT                           ; next column
        brne    IN_SNES_2_LOOP
        cbr     SNES_FLAG,1<<SNES_F_2ND         ; 2nd part of work is done
        sbr     SNES_FLAG,1<<SNES_F_DATA        ; new SNES key data is available now
IN_SNES_2_END:
        ret



; do step actions
;  - for i = 0..6:
;     - in step i^2: output black/white picture, turn on pixel with value > i
;  - in step 42: read SNES keys (1st half)
;  - in step 46: read SNES keys (2nd half)
;  - uses: X, SNES_FLAG, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 2
STEP_ACTIONS:
        cpi     STEP,10
        brsh    STEP_ACTIONS_10
        cpi     STEP,3
        brsh    STEP_ACTIONS_3
; in step 0: output black/white picture, turn on pixel with value >= 1
; in step 1: output black/white picture, turn on pixel with value >= 2
        cpi     STEP,0
        brne    STEP_ACTIONS_NOT_0
        rjmp    OUT_BW_1
STEP_ACTIONS_NOT_0:
        cpi     STEP,1
        brne    STEP_ACTIONS_NOT_1
        rjmp    OUT_BW_2
STEP_ACTIONS_NOT_1:
        ret
STEP_ACTIONS_3:
; in step 4: output black/white picture, turn on pixel with value >= 3
; in step 9: output black/white picture, turn on pixel with value >= 4
        cpi     STEP,4
        brne    STEP_ACTIONS_NOT_4
        rjmp    OUT_BW_3
STEP_ACTIONS_NOT_4:
        cpi     STEP,9
        brne    STEP_ACTIONS_NOT_9
        rjmp    OUT_BW_4
STEP_ACTIONS_NOT_9:
        ret
STEP_ACTIONS_10:
        cpi     STEP,30
        brsh    STEP_ACTIONS_30
; in step 16: output black/white picture, turn on pixel with value >= 5
; in step 25: output black/white picture, turn on pixel with value >= 6
        cpi     STEP,16
        brne    STEP_ACTIONS_NOT_16
        rjmp    OUT_BW_5
STEP_ACTIONS_NOT_16:
        cpi     STEP,25
        brne    STEP_ACTIONS_NOT_25
        rjmp    OUT_BW_6
STEP_ACTIONS_NOT_25:
        ret
STEP_ACTIONS_30:
; in step 36: output black/white picture, turn on pixel with value >= 7
; in step 42: read SNES keys (1st half)
; in step 46: read SNES keys (2nd half)
        cpi     STEP,36
        brne    STEP_ACTIONS_NOT_36
        rjmp    OUT_BW_7
STEP_ACTIONS_NOT_36:
        cpi     STEP,42
        brne    STEP_ACTIONS_NOT_42
        rjmp    IN_SNES_1
STEP_ACTIONS_NOT_42:
        cpi     STEP,46
        brne    STEP_ACTIONS_NOT_46
        rjmp    IN_SNES_2
STEP_ACTIONS_NOT_46:
        ret



; timer interrupt
;  - every 192 us (2304 cycles)
; uses: X (reserved for interrupt), I_SREG, I_TMP, I_CNT, I_REG0, I_REG1, I_REG2, STEP
;  - used stack depth (without call to this function): 4
INT_TIMER:
        in      I_SREG,SREG
; do step actions
        rcall   STEP_ACTIONS
; increment output step number
        inc     STEP
        cpi     STEP,49
        brlo    INT_TIMER_STEP_LOWER_49
        clr     STEP
INT_TIMER_STEP_LOWER_49:
        out     SREG,I_SREG
        reti



; flip output buffers
;  - exchanges the pointer to the output buffers
;  - received output values will be output from now on
;  - restarts serial reception (i.e. discards current frame)
;  - uses: RECV_STEP, TMP
;  - used stack depth (without call to this function): 0
OUT_FLIP:
; start receiving new serial frame
;  - neccessary because output buffers will be changed
        clr     RECV_STEP
; exchange output buffers
        cpi     OUT_PTR_FLAG,0
        brne    OUT_FLIP_X
; flag 0: flag 1, output buffer 2 for output, output buffer 1 for receiving
        cli                                     ; turn off interrupts
        ldi     OUT_PTR_FLAG,1
        ldi     TMP,high(OUT_BUF_1+18*3)
        mov     OUT_PTR_OUT_E_H,TMP
        ldi     TMP,low(OUT_BUF_1+18*3)
        mov     OUT_PTR_OUT_E_L,TMP
        ldi     TMP,high(OUT_BUF_2)
        mov     OUT_PTR_RECV_H,TMP
        ldi     TMP,low(OUT_BUF_2)
        mov     OUT_PTR_RECV_L,TMP
        sei                                     ; turn on interrupts
        ret
; flag != 0: flag 0, output buffer 1 for output, output buffer 2 for receiving
OUT_FLIP_X:
        cli                                     ; turn off interrupts
        ldi     OUT_PTR_FLAG,0
        ldi     TMP,high(OUT_BUF_2+18*3)
        mov     OUT_PTR_OUT_E_H,TMP
        ldi     TMP,low(OUT_BUF_2+18*3)
        mov     OUT_PTR_OUT_E_L,TMP
        ldi     TMP,high(OUT_BUF_1)
        mov     OUT_PTR_RECV_H,TMP
        ldi     TMP,low(OUT_BUF_1)
        mov     OUT_PTR_RECV_L,TMP
        sei                                     ; turn on interrupts
        ret



; image data for BM picture
OUT_BM_DATA:
        .db     0x00, 0x00, 0x00, 0x7F, 0x7F, 0x7F
        .db     0x7F, 0xFF, 0x7F, 0x49, 0xFF, 0x49
        .db     0x49, 0xDB, 0x49, 0x36, 0xB6, 0x36
        .db     0x36, 0x7E, 0x36, 0x00, 0x6C, 0x00
        .db     0x00, 0x00, 0x00, 0x7F, 0x7F, 0x7F
        .db     0x7F, 0xFF, 0x7F, 0x02, 0xFE, 0x02
        .db     0x04, 0x04, 0x04, 0x04, 0x0C, 0x04
        .db     0x02, 0x0A, 0x02, 0x7F, 0x7F, 0x7F
        .db     0x7F, 0xFF, 0x7F, 0x00, 0xFE, 0x00



; show BM picture
;  - uses: Y, Z, CNT, TMP
;  - used stack depth (without call to this function): 0
OUT_BM:
        ldi     ZH,high(2*OUT_BM_DATA)          ; load base address of image ("2*" to convert word address to byte address)
        ldi     ZL,low(2*OUT_BM_DATA)
        mov     YH,OUT_PTR_RECV_H               ; get address of output buffer to put new data in
        mov     YL,OUT_PTR_RECV_L
        ldi     CNT,18*3
OUT_BM_LOOP:
        lpm     TMP,Z+                          ; read byte from program memory
        st      Y+,TMP                          ; write byte to output buffer
        dec     CNT
        brne    OUT_BM_LOOP
        rjmp    OUT_FLIP                        ; show new picture



; serial reception
; uses: Z, CNT, TMP, REG0, REG1, RECV_STEP
;  - used stack depth (without call to this function): 0
RECV:
; check for reception
        sbis    UCSRA,RXC
        rjmp    RECV_END                        ; nothing received
; check for errors
        sbic    UCSRA,FE
        rjmp    RECV_ERR                        ; frame error
        sbic    UCSRA,DOR
        rjmp    RECV_ERR                        ; data overrun
; check if behind message
        cpi     RECV_STEP,12+18*8
        brlo    RECV_STEP_OK
        clr     RECV_STEP                       ; behind message -> start receiving a new message
RECV_STEP_OK:        
; check if in header or in data part
        cpi     RECV_STEP,12
        brlo    RECV_HEADER
        rjmp    RECV_DATA

; receive error
RECV_ERR:
        in      TMP,UDR                         ; read data from USART to clear error
        clr     RECV_STEP                       ; start receiving a new message
RECV_END:
        ret

; expected header data (MCUF header 18x8-1/256)
RECV_HEADER_DATA:
        .db     0x23, 0x54, 0x26, 0x66          ; magic (first byte is also contained in code)
        .db     0x00, 0x08                      ; height: 8
        .db     0x00, 0x12                      ; width: 18
        .db     0x00, 0x01                      ; channels: 1
        .db     0x00, 0xFF                      ; maxval: 255

; received a header byte
RECV_HEADER:
        in      REG0,UDR                        ; fetch data byte from USART
        ldi     ZH,high(2*RECV_HEADER_DATA)     ; load base address of header data ("2*" to convert word address to byte address)
        ldi     ZL,low(2*RECV_HEADER_DATA)
        clr     TMP                             ; add offset
        add     ZL,RECV_STEP
        adc     ZH,TMP
        lpm     REG1,Z                          ; fetch expected data byte from program memory
        cp      REG0,REG1                       ; compare with received data byte
        breq    RECV_HEADER_OK
        clr     RECV_STEP                       ; mismatch -> start receiving new message
        ldi     TMP,0x23                        ; check if first byte of message (first byte of magic)
        cp      REG0,TMP
        breq    RECV_HEADER_OK                  ; first byte of message/magic -> accept it
        ret
RECV_HEADER_OK:
        inc     RECV_STEP                       ; match -> next byte
        clr     RECV_X_3                        ; next pixel to receive is topleft pixel
        clr     RECV_Y_MASK                     ;  - only needed for last header byte, but faster and more easy this way
        inc     RECV_Y_MASK
        ret

; received a data byte
RECV_DATA:
        in      REG0,UDR                        ; fetch data byte from USART
        mov     ZH,OUT_PTR_RECV_H               ; get base address of output buffer to put received data in
        mov     ZL,OUT_PTR_RECV_L
        clr     TMP                             ; get destination address (dest_addr := base_addr + x_coord * 3)
        add     ZL,RECV_X_3
        adc     ZH,TMP
        mov     REG1,RECV_Y_MASK                ; get negated bit mask into REG1 (mask = 1 << y_coord; neg_mask := ~mask)
        com     REG1
        ld      TMP,Z                           ; store D0 of pixel value (D5 of received data byte)
        and     TMP,REG1
        sbrc    REG0,5
        or      TMP,RECV_Y_MASK
        st      Z+,TMP
        ld      TMP,Z                           ; store D1 of pixel value (D6 of received data byte)
        and     TMP,REG1
        sbrc    REG0,6
        or      TMP,RECV_Y_MASK
        st      Z+,TMP
        ld      TMP,Z                           ; store D2 of pixel value (D7 of received data byte)
        and     TMP,REG1
        sbrc    REG0,7
        or      TMP,RECV_Y_MASK
        st      Z+,TMP
        inc     RECV_STEP                       ; next byte
        ldi     TMP,3                           ; next pixel
        add     RECV_X_3,TMP
        ldi     TMP,3*18
        cp      RECV_X_3,TMP
        brlo    RECV_NEXT_END
        clr     RECV_X_3                        ; next line
        lsl     RECV_Y_MASK
RECV_NEXT_END:
        cpi     RECV_STEP,12+18*8               ; end of frame - output picture
        brsh    RECV_DATA_OUTPUT
        ret

; output picture
RECV_DATA_OUTPUT:
; reset timeout timer
        cli                                     ; turn off interrupts
        clr     TMP                             ; reset timer 1 (timeout timer)
        out     TCNT1H,TMP
        out     TCNT1L,TMP
        in      TMP,OCF1A                       ; clear timeout flag (timer 1 compare match)
        sbr     TMP,1<<OCF1A
        out     TIFR,TMP
        sei                                     ; turn on interrupts
; show received data
        rjmp    OUT_FLIP                        ; show received data



; serial tranmission
; uses: Z, SNES_FLAG, TRANS_STEP, CNT, TMP
;  - used stack depth (without call to this function): 0
TRANS:
        sbrc    SNES_FLAG,SNES_F_DATA           ; check for new data to transmit
        rjmp    TRANS_DO
        clr     TRANS_STEP                      ; no new data -> restart with byte 0 if new data arrives
        ret
TRANS_DO:
        sbis    UCSRA,UDRE                      ; exit if no space in USART transmit buffer
        ret
        clr     REG0                            ; get SNES mask (1st key) into REG0 (mask1 := 1 << (step / 2) & 3)
        inc     REG0
        sbrc    TRANS_STEP,2
        lsl     REG0
        sbrc    TRANS_STEP,2
        lsl     REG0
        sbrc    TRANS_STEP,1
        lsl     REG0
        mov     REG1,REG0                       ; get SNES mask (2nd key) into REG1 (mask2 := mask1 << 4)
        swap    REG1
        sbrc    TRANS_STEP,0                    ; transmitting high or low byte? (highbyte is transmitted first)
        rjmp    TRANS_LOW
        ldi     ZH,high(SNES_BUF+6)             ; get address of "key 12/13" in SNES key buffer
        ldi     ZL,low(SNES_BUF+6)
        clr     REG2                            ; get data to transmit (SNES buttons 7..11)
        ldi     CNT,3
TRANS_HIGH_LOOP:
        lsl     REG2
        ld      TMP,-Z                          ;  - load 2 keys
        and     TMP,REG1                        ;  - 2nd key
        brne    TRANS_HIGH_BIT0                 ;  - SNES key bits are inverted
        inc     REG2
TRANS_HIGH_BIT0:
        lsl     REG2
        ld      TMP,Z                           ;  - load 2 keys again
        and     TMP,REG0                        ;  - 1st key
        brne    TRANS_HIGH_BIT00                ;  - SNES key bits are inverted
        inc     REG2
TRANS_HIGH_BIT00:
        dec     CNT
        brne    TRANS_HIGH_LOOP
        cpi     TRANS_STEP,0                    ; set sync bit (D7) if first byte of first SNES pad
        brne    TRANS_FINISH
        ldi     TMP,0x80
        or      REG2,TMP
        rjmp    TRANS_FINISH
TRANS_LOW:
        ldi     ZH,high(SNES_BUF+3)             ; get address of key 6/7 in SNES key buffer
        ldi     ZL,low(SNES_BUF+3)
        clr     REG2                            ; get data to transmit (SNES buttons 0..6)
        ldi     CNT,3
TRANS_LOW_LOOP:
        lsl     REG2
        ld      TMP,-Z                          ;  - load 2 keys
        and     TMP,REG1                        ;  - 2nd key
        brne    TRANS_LOW_BIT0                  ;  - SNES key bits are inverted
        inc     REG2
TRANS_LOW_BIT0:
        lsl     REG2
        ld      TMP,Z                           ;  - load 2 keys again
        and     TMP,REG0                        ;  - 1st key
        brne    TRANS_LOW_BIT00                 ;  - SNES key bits are inverted
        inc     REG2
TRANS_LOW_BIT00:
        dec     CNT
        brne    TRANS_LOW_LOOP
TRANS_FINISH:
        out     UDR,REG2                        ; transmit data
        inc     TRANS_STEP                      ; next step
        cpi     TRANS_STEP,8                    ; finished?
        brlo    TRANS_END
        cbr     SNES_FLAG,1<<SNES_F_DATA        ; SNES data is sent
        clr     TRANS_STEP                      ; restart with first byte next time
TRANS_END:
        ret



; main program - initialization
;  - used stack depth: 2
MAIN:
        wdr                                     ; watchdog reset
        clr     STEP                            ; output step counter to first output step
        clr     OUT_PTR_FLAG                    ; initialize output buffer pointers
        rcall   OUT_FLIP
        rcall   OUT_BM                          ; show BM picture
        clr     SNES_FLAG                       ; no new SNES key data available
        clr     RECV_STEP                       ; start receiving with a new message
        clr     TRANS_STEP                      ; start transmitting with first byte
        clr     TMP                             ; reset timer 1 (timeout timer)
        out     TCNT1H,TMP
        out     TCNT1L,TMP
        in      TMP,OCF1A                       ;  - clear timeout flag (timer 1 compare match)
        sbr     TMP,1<<OCF1A
        out     TIFR,TMP
        sei                                     ; turn on interrupts



; main program - main loop
MAIN_LOOP:

; watchdog reset
        wdr

; receive serial data
        rcall   RECV

; transmit serial data
        rcall   TRANS

; timeout -> BM picture
        in      TMP,TIFR                        ; skip if timeout flag (timer 1 compare match) is set
        sbrs    TMP,OCF1A
        rjmp    MAIN_TIMEOUT_END
        rcall   OUT_BM                          ; show BM picture
        cli                                     ; turn off interrupts
        in      TMP,OCF1A                       ; clear timeout flag (timer 1 compare match)
        sbr     TMP,1<<OCF1A
        out     TIFR,TMP
        sei                                     ; turn on interrupts
MAIN_TIMEOUT_END:

; bottom of main loop
        rjmp     MAIN_LOOP

